/*
 *  (C) Copyright ECWMF http://www.ecmwf.int
 *
 *  This software is licensed under the terms of the Apache Licence Version 2.0
 *  which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 *  In applying this licence, ECMWF does not waive the privileges and immunities
 *  granted to it by virtue of its status as an intergovernmental organisation nor
 *  does it submit to any jurisdiction.
 *
 *  Filterable list
 *  ---------------
 *  A list of objects that has filters applied.
 *  Use the subscribe method to be notified of filtering changes.
 *  Use the have_filtering promise to ensure that the list creation is complete.
 *
 *  To have the list be sorted, set the sort_by property of the list. This will
 *  assume that each list item has a key with this name.
 */
define(['jquery', 'observable', 'throttle-debounce'], function($, Observable) {

var list = $.extend(new Observable(), {

	items: null,    // all items
	filtered: null, // array of indexes to the full list.

	filters: [],

	// TODO rename these to added, removed
	changed: {
		to_show: [],
		to_hide: []
	},

	selection: $.extend(new Observable(), {always: true}),
	on: {
		items: new Observable() // indicates list contents have completely changed, eg change of chart package.
	},

	set_collection: function(collection, done) {
		var self = this;
		list.collection = collection;
		list.collection.fetch(function(items){
			list.have_items.resolve(list.items = items);
			self.on.items.emit();
			done && done();
		});
	},

	// TODO reduce the debounce time. Following changes it may no longer be needed.
	filter: $.debounce(200, function() {
		var old_len = list.filtered ? list.filtered.length : -1 // -1 ensures change is emitted first time
		list.changed = {to_hide: [], to_show: []};
		list.filtered = [];

		if(!list.items) return console.warn('empty list');

		var n_matches = 0;
		list.each(function(v, i){
			var visible = this.filter_pass;

			if((this.filter_pass = list.filter_item(this))){
				if(!visible) list.changed.to_show.push(this);
				list.filtered.push(i);
				this.filtered_idx = n_matches++;
			}
			else{
				if(visible) list.changed.to_hide.push(this);
			}

			// TODO if this turns out to be useful it should be optimised and managed by the facet widget.
			this.filter_pass_no_facets = list.filter_item_without_facets(this);
		});

		if(list.sort_by){
			list.filtered.sort(function(a, b){
				return list.items[a].data[list.sort_by] > list.items[b].data[list.sort_by];
			});
		}

		// only emit if the list has changed  - currently we only check the empty case
		(!old_len && !list.filtered.length) || list.emit();

		// fix selection if invalid
		if($.inArray(+list.selection.value, list.filtered) < 0){
			list.selection.set((list.selection.always && list.filtered.length) ? list.filtered[0] : null);
		}

		list.have_filtering.resolve();
	}),

	/*
	 * return true or false
	 */
	filter_item: function(item) {
		for(var i in list.filters){
			if((list.filters[i].disabled !== true) && !list.filters[i].fn(item)) return false;
		}
		return true;
	},

	filter_item_without_facets: function(item) {
		for(var i in list.filters){
			if((list.filters[i].name !== 'facets' && list.filters[i].disabled !== true) && !list.filters[i].fn(item)) return false;
		}
		return true;
	},

	lookup_filter_by_name: function(name) {
		for(var i in list.filters){
			var f = list.filters[i];
			if(f.name === name) return f;
		}
	},

	lookup_idx_by_name: function(name) {
		for(var i in list.items){
			if(list.items[i].name() === name) return i;
		}
	},

	remove_filter: function(name) {
		var filter = this.lookup_filter_by_name(name);
		list.filters = $.map(list.filters, function(v){ if(v.name !== name) return v; });
	},

	have_items: $.Deferred(),
	have_filtering: $.Deferred(),

	each: function(fn) {
		// TODO why is i a string!?
		for(var i in list.items) fn.call(list.items[i], list.items[i], +i);
	},

	each_by_facet: function(facet, fn1, fn2) {
		var self = this;
		var g = -1, ig = 0, gr = {};
		var other = {name: 'Other'};
		for(var i=0;i<list.filtered.length;i++){
			var item = self.items[list.filtered[i]];
			var group_name = (item.data.tags && item.data.tags[facet] || [])[0];
			if(!i || (group_name && group_name !== gr.name)){
				g++;
				gr.name = group_name || other.name;
				fn1(gr.name);
				ig = 0;
			}
			fn2(item, i, ig++, group_name, g);
		}
	}
});

return list;

});
