/*
 *  ECWMF Products
 *  --------------
 *
 *  A further (2014) attempt at a generic collection object for chart products.
 *
 *  Please note that this conforms to the Product Collection Interface (see documentation). Do not modify the API.
 *
 */
define(['config', 'jquery', 'ajax_queue', 'storage', 'bus', 'utils', 'throttle-debounce'], function(ECMWF, $, Ajaxqueue, localStorage, Bus){

	/*
	 * Facets are formatted as below:
	 * 
	 * product: {
	 *     tags: {
	 *         facet_name1: value1,
	 *         facet_name2: [value2, value3]
	 *     }
	 * }
	 *
	 */

	var in_dev = true; // TODO

	var settings = {
		score: ECMWF.options.product_scoring || false,   // reject misbehaving products
		cache_expire_mins: ECMWF.options.thumbnail_cache || 10
	}

	var ChartListItem = function(item) {
		if(!item) return console.warn('ChartListItem: no data');

		$.extend(this, mixins);

		this.data = item;

		// tags are forced to be arrays for efficient filtering.
		for(var tags in item.tags){
			var t = item[tags];

			if(t) this.data.tags[tags] = [t]; // for 'title'. can probably be remove later.
			if(typeof item.tags[tags] === 'string') this.data.tags[tags] = [item.tags[tags]];
		}

		this.thumbnail = function(args, done) {
			done = done || $.noop;

			var tn = this.cache.thumbnail.call(this);
			if(tn){
				return done(tn);
			}

			Thumbnail_queue.add(this, args, done);
		}

		this.cache = {
			thumbnail: function(val) {
				return this.cache._prop('thumbnail', val);
			},

			score: function(val) {
				return settings.score
					? (this.cache._prop('score', val) || 0) // TODO is the zero needed?
					: 0;
			},

			_prop: function(name, val) {

				this._item = this._item || products._cache(this.product.name()) || this._empty();
				if(val){
					this._item[name] = val;
					products.store(this.product.name(), this._item);
				}else{
					products.expired(this._item) && this._clear();
				}

				if(this._item[name]){
					return this._item[name];
				}
			},

			_empty: function() {
				return this._item = {timestamp: new Date().getTime()};
			},

			_clear: function() {
				this._empty();
				products.store(this.product.name(), this._item);
			},

			_item: null,
			product: this
		}
	}

	ChartListItem.prototype = {
		name: function() {
			return this.data.name;
		},

		title_: function() {
			return this.data.title;
		},

		long_title: function() {
			return this.data.long_title || this.data.title;
		},

		keywords: function(order) {
			return '';
		},

		has_widget: function(widget) {
			return !!~$.inArray(widget, this.data.widgets)
		}
	}

	// In the absence of generic ajax bundling, thumbnail requests are bundled together here
	Thumbnail_queue = {
		max_bundle: 1,
		items: [],
		in_progress: [],
		timer: null,
		add: function(item, args, done) {
			var self = this, q = Thumbnail_queue, lists;

			for(var ai in lists = [q.items, q.in_progress]){
				for(var i in lists[ai]){
					var l = lists[ai][i];
					if(item.name() === l.item.name()){
						return l.done.push(done);
					}
				}
			}

			q.items.push({item: item, args: args, done: [done]});
			q.timer = q.timer || setInterval(function() {
				if(q.items.length){
					do q._fetch(q.items.splice(0, q.max_bundle)); while(q.items.length >= q.max_bundle);
				}else{
					q.timer = clearInterval(q.timer);
				}
			}, 100);
		},
		_fetch: function(items) {
			var self = this, q = Thumbnail_queue;

			q.in_progress = q.in_progress.concat(items);

			var args = items[0].args; // args (eg, size) are assumed to be common to all requests
			var first = items[0].item;
			var names = $.map(items, function(v){ return v.item.name() });

			function _fail(error) {
				$.each(items, function(i, item){
					var product = item.item;
					settings.score && product.cache.score.call(product, product.cache.score.call(product) - 1);
					$.each(item.done, function(i, done){ done(null, error || 'failed'); });
				});
			}

			Ajaxqueue($.extend(
				{
					url        : '/backend/product-thumbnail-inline/',
					json       : {name: names},
					tag        : 'grid-' + names.join(','),
					priority   : products.thumbnail_priority,
					logErrors  : 'Get thumbnail (' +  first.title_() + ')',
					complete: function(result){
						var indexes = [];
						for(var i in items){
							var item = items[i].item
							for(var j in q.in_progress){
								var l = q.in_progress[j];
								if(l.item.name() === item.name()){
									indexes.push(+j)
									break
								}
							}
						}
						indexes = indexes.sort(function(a, b){ return b - a });
						for(i in indexes){
							q.in_progress.splice(indexes[i], 1);
						}

						$.each(items, function(i, item) {
							var product = item.item;
							if(product.validate([
								['has result', function(){ return result; }],
								['has url', function(){ return result.thumbnails || result.url; }]
							])){
								product.cache.thumbnail.call(product, result.url)
								$.each(item.done, function(j, done){ done(result.thumbnails[i].url, result.thumbnails[i].img); });
							}
							else _fail('validation failed');
						});
					},
					error: _fail
				},
				args)
			);
		}
	}

	ECMWF.options.timeout && $.extend(Ajaxqueue.options, {
		timeout: ECMWF.options.timeout * 1000
	});

	var mixins = {
		validate: function(tests) {
			for(var t in tests) if(!tests[t][1]()){
				var message = tests[t][0];
				console.warn('validate: "' + message + '" failed.');
				this.error_msg = "test '" + message + "' failed";
				return false;
			}
			return true;
		}
	}

	var products = new (function Products(){

		$.extend(this, mixins);

		this.query = {}
		this.thumbnail_priority = 'low',

		this.set_package = function(pkg){
			this.products = null;
			this._facets = null;
			$.extend(this.query, {package: pkg})
		}

		this.fetch = function(query, on_success){
			var self = this;
			if(!on_success && $.isFunction(query)){ on_success = query; query = null; }

			if(self.products) return on_success(self.products);

// why not using a promise? TODO make products.products a promise
			self.__fetch_callbacks = self.__fetch_callbacks || [];
			self.__fetch_callbacks.push(on_success);
			if(self.__fetch_callbacks.length > 1) return;

			function have_products(result) {
				__DEV__ && Bus().send('event-start', ['start verification']);
				window.is_node || console.log('products.fetch: ', result);
				self.products = $.map(result.records, function(item){
					if(self.validate([
							['not empty', function(){ return item; }],
							['has name', function(){ return item.name; }],
							['has title', function(){ return item.title; }],
							['has no empty facets', function(){
								if(item.tags){
									for(var tags in item.tags){
										if($.isArray(item.tags[tags]) && !item.tags[tags].length){
											console.warn('product.validate: product error: ' + item.name + ': tag is empty array (' + tags + ')');
											delete item.tags[tags];
										}
									}
								}
								return true;
							}],
							['path has no trailing slash', function(){ if($.isArray(item.path))
								for(var i in item.path){
									var p = item.path[i];
									if((item.path[i] = item.path[i].replace(/\/$/, '')) !== p)
									console.warn('path has trailing slash: ', p, item.name);
								}
								return true;
							}],
							['has keywords', function(){ return item.keywords || (item.keywords = ' '); }] // temporary. should be server-side.
						])){
						var product = new ChartListItem(item);
						if((!in_dev) || (product.cache.score.apply(product) > -1)) return product;
					}
				});
				if(self.products.length){
					console.log('n_products=', self.products.length, ' (', (result.query_count - self.products.length), ' rejected)')
				}else{
					console.log('no products found for:', query || self.query)
				}

				self.packages.then(function() {
					if(self.__fetch_callbacks){
						while(self.__fetch_callbacks.length){
							var callback = self.__fetch_callbacks.shift();
							callback(self.products);
						}
						delete self.__fetch_callbacks;
					}
				});
			}

			if(window._ECMWF && window._ECMWF.products){
				if(_ECMWF.all_products){
					$.each(_ECMWF.all_products, function(i, v){
						var p = $.findWhere(_ECMWF.products, {name: v.name})
						if(p){
							p.thumbnail = v.thumbnail
						}
					})
				}
				have_products({records: window._ECMWF.products});
			}else{
				Ajaxqueue({
					url: '/plots/find-product/',
					json: $.extend({}, {query: query || self.query}),
					complete: have_products
				});
			}
		}

		/*
		 *  Generate a list of facets from the product list.
		 *  The method is synchronous and relies only on product data, not package data.
		 */
		this.facets = function(){
			var self = this;
			if(!this._facets){
				this._facets = {};

				$.each(this.products, function(i, product){
					if(product.data.tags){
						$.each(product.data.tags, function(i, v){

							self._facets[i] || (self._facets[i] = {});

							if($.isArray(v))
								$.each(v, function(j, v){ self._facets[i][v] = true; });
							else
								self._facets[i][v] = true;
						});
					}
				});
				$.each(self._facets, function(i, v){
					self._facets[i] = $.map(v, function(_, i){ return i; }).sort();
				});
			}

			return this._facets;
		}

		/*
		 * Now that products can be in more than one package, the pkg needs to be specified
		 */
		this.sorted_facets = function(pkg, blacklist, done) {
			var self = this;
			if(arguments.length === 2){
				done = blacklist;
				blacklist = [];
			}

			this.products.length && this.packages.then(function() {
				self.facets();

				var sorted = pkg.facets || [];

				var keys = sorted.length
					? $.map(self._facets, function(i, key){ if(!~$.inArray(key, blacklist)) return {name: key, values: i, order: $.inArray(key, sorted)}; })
					: $.map(self._facets, function(i, key){ if(!~$.inArray(key, blacklist)) return {name: key, values: i, order: 0}; });

				done(keys.sort(function(a, b) {
					return a.order - b.order;
				}));
			});
		}

		this.packages = $.Deferred(function(df){
			console.log('products.packages');
			var self = this;

			Ajaxqueue({
				url: '/plots/get-packages/',
				priority: 'very-high',
				complete: function(result){
					if(result.packages){
						products._packages = result.packages;
						products.hierarchy.update(result.categories);
						df.resolve(products._packages);
					}
				}
			});
		}).promise()

		this.hierarchy = {
			products: this,
			tree: {},

			update: function(categories) {
				var self = this;
				/*
				this.tree = {};
				$.each(this.products.products, function(i, p){
					for(var k in p.data.path){
						var path = p.data.path[k].split('/');
						var cursor = self.tree;
						for(var j in path){
							if(path[j]){
								cursor = cursor[path[j]] = cursor[path[j]] || {};
							}
						}
					}
				});
				*/
				this.tree = categories;

				// fill in missing categories TODO move to server.
				for(var i in this.products._packages){
					var pkg = this.products._packages[i];
					if(!$.findWhere(this.tree, {name: pkg.name})){
						this.tree.push({name: pkg.name, title: pkg.title});
					}
				}
			},

			children: function(path) {
				var node = this._child(path, 0, this.tree);
				return node ? node.children : [];
			},

			lookup: function(path) {
				if(!path || path === 'all') return {title: 'ECMWF charts'}; // special case because the tree does not currently have a root item

				$.isEmptyObject(this.tree) && console.warn('hierarchy.lookup: no package list');
				if(!$.isArray(path)) path = path.replace(/\/#.*/, '').split('/');
				return this._child(path, 0, this.tree);
			},

			_child: function(path, p, tree){
				for(var i in tree){
					var node = tree[i];
					if(node.name === path[p]){
						if(p === path.length - 1){
							return node;
						}else{
							p++;
							return this._child(path, p, node.children);
						}
					}
				}
			},

			path_is_valid: function(path) {
				return true;
			}
		}

		this.product_by_name = function(name) {
			for(var i in this.products){
				if(this.products[i].data.name === name) return this.products[i];
			}
		}

		this.cancel_thumbnails = function() {
			//Ajaxqueue.cancelByTag(new RegExp(self.eventNamespace + '$')); - does not work when batching requests
			this._cancel_warning_done || console.log('products.cancel_thumbnails TODO');
			this._cancel_warning_done = true;
		}

		/*
		 *  Get html text content associated with the given product or category.
		 */
		this.content = function(path, done) {
			if(!path) return; // TODO drupal module does not currently support text for the landing page

			ECMWF.options.ajax.chart_content && Ajaxqueue({
				url: ECMWF.options.ajax.chart_content + '/' + path,
				name: 'content',
				complete: done
			}) || done();
		}

		/*
		 *  Get text content temporarily stored in the product database
		 */
		this.content_temp = function(product_name, done) {
			Ajaxqueue({
				url: ECMWF.options.ajax.plots.get_product_text,
				data: JSON.stringify({product: {name: product_name}}),
				success: done
			});
		}

		this._cache =
			localStorage ? function(name) {
				if(!this.__cache){
					this.__cache = JSON.parse(localStorage.products || '{}');
				}
				return name ? this.__cache[name] : this.__cache;
			}
			: function(){ return {}; }

		this._store = $.debounce(2000, function() {
			localStorage.products = JSON.stringify(this.__cache);
		})

		this.store = localStorage ? function(name, item) {
			this.__cache[name] = item;
			this._store();
		} : $.noop;

		this.expired = function(item) {
			return item && ((!item.timestamp) || (new Date().getTime() - item.timestamp > 1000 * 60 * settings.cache_expire_mins));
		}

		setTimeout(function clean_cache() {
			$.each(products._cache(), function(i, item) {
				if(products.expired(item)){
					delete products._cache()[i]
					products._store();
				}
			});
			//console.log('products.clean_cache: size=', ECMWF.n_properties(products._cache()));
		}, 5000);

		// the methods below from the previous version have not yet been re-implemented

		/*
		this.keywords = function(order){
			console.warn('products.keywords FN NOT USED?');
			var self = this;

			function sort_keywords_by_popularity(keywords){
				function sortfunction(a, b){
					return a.n < b.n;
				}
				var arr = [];
				for(var i in keywords){
					arr.push({name:i, n: keywords[i]});
				}
				arr.sort(sortfunction);
				return arr;
			}

			var tags = {};
			if(true || order === 'by_popularity'){
				self.list.each(function(){
					if(this.keywords){
						var _tags = this.keywords.split(' '); //warning different separator to eccharts
						for(var i in _tags){
							if(_tags[i]){
								if(typeof tags[_tags[i]] === 'undefined') tags[_tags[i]] = 1;
								else tags[_tags[i]]++;
							}
						}
					}
				});
				return sort_keywords_by_popularity(tags);
			}

			//for unsorted tags, return an object not an array.
			//-this path is currently not used.
			self.list.each(function(){
				var _tags = this.keywords.split(', ');
				for(var i in _tags){
					if(_tags[i]){
						if(typeof tags[_tags[i]] === 'undefined') tags[_tags[i]] = 1;
						else tags[_tags[i]]++;
					}
				}
			});
			return tags;
		},
		*/
	})();

	return products;
});
